; *****************************************************************************
; *****************************************************************************
;
;       File    : Display.Asm
;       Author  : Paul Robson (autismuk@aol.com)
;       Purpose : Render the Display
;       Created : 7th October 1997
;       Modified: 12th November 1997
;
;       Functions :
;               SCRebuild               Rebuild the Entire Screen
;               SCInitialise            Initialise the Screen
;               SCSetSync               Set the Sync Flag
;               SCVSync                 Do the Vertical Sync Code
;               SCExecDisplayFile       Handle Code at >= $8000
;               SCFastPaint             Fast Display Painter
;               SCResetBackground       Reset the Background
;
; *****************************************************************************
; *****************************************************************************

.386p 
.model flat
include extern.inc

.data
VRAMPtr         dd      0               ; Video RAM Pointer
VideoAddress    dd      0               ; Current Video Line Address
RenderPointer   dd      0               ; rendering address
InFastMode      dd      0               ; fast mode blank test
_OldVSync       db      0,0,0,0         ; Last VSync value
_LastChange     dd      0               ; NMI Count of last change
.code

; *****************************************************************************
;
;                   Set Vertical Sync to AL (0 = Off,1 = On)
;
; *****************************************************************************

SCSetSync Proc
        cmp     [_OldVSync],al          ; VSYNC changed
        je      _SCSS0                  ; if not,exit.

        mov     [_OldVSync],al          ; update old vertical sync
        or      al,al                   ; skip VSync action if switching on
        jnz     _SCSS1

        mov     eax,[NMICounter]        ; VSYNC must be at least 4 NMIs
        sub     eax,[_LastChange]       ; to sync correctly.
        cmp     eax,4                   ; a low pulse of 4 NMI's will
        jb      _SCSS1                  ; force beam to the top.

        call    SCVSync                 ; do the VSync code
        inc     [VSyncCount]            ; increment the vsync counter
        mov     [NMIVSCounter],0        ; reset since-VSYNC NMI Counter
        mov     [Cycles],0
     
_SCSS1: mov     eax,[NMICounter]        ; reset the counter flag
        mov     [_LastChange],eax
_SCSS0: ret

EndP SCSetSync

; *****************************************************************************
;
;                           Come here to do a VSYNC
;
; *****************************************************************************

SCVSync Proc
        cmp     [Redraw],0              ; Screen been redrawn ?
        jnz     _SCV1                   ; on this cycle

        call    SCRebuild               ; Copy Physical to Logical
_SCV1:
        inc     [Redraw]                ; Adjust redraw parameters
        and     [Redraw],3              ; for next cycle
        mov     [InFastMode],0          ; not in fast now
        mov     [VideoLine],0           ; reset current line to zero
        mov     eax,offset VBuffer      ; start address for drawing
        add     eax,320*3+32            ; centre it
        mov     [VideoAddress],eax
        ret
EndP SCVSync

; *****************************************************************************
;
;          Initialise Sinclair Display. On entry EAX = VGA Address
;
; *****************************************************************************

SCInitialise Proc
        mov     [VRAMPtr],eax           ; save the video memory address
        ret
EndP SCInitialise

; *****************************************************************************
;
;               Draw the screen border (non display space)
;
; *****************************************************************************

SCFrame Proc
        mov     eax,[Background]        ; fill VBuffer with background
        cmp     [FrmRate],9
        jne     _SCF1
        mov     eax,04040404h           ; or red if in Turbo mode
_SCF1:  mov     ebx,offset VBuffer
        mov     ecx,320*200/4
_SCF0:  mov     [ebx],eax
        add     ebx,4
        dec     ecx
        jnz     _SCF0
        ret
EndP SCFrame

; *****************************************************************************
;
;                        Rebuild the entire screen
;
; *****************************************************************************

SCRebuild Proc

        mov     ebx,offset VBuffer      ; copy logical to physical
        mov     ecx,[VRAMPtr]
        mov     edx,320*200/4/4

_SCR0:  mov     eax,[ebx]               ; do it 16 bytes per loop
        mov     [ecx],eax
        mov     eax,[ebx+4]
        mov     [ecx+4],eax
        mov     eax,[ebx+8]
        mov     [ecx+8],eax
        mov     eax,[ebx+12]
        mov     [ecx+12],eax

        add     ebx,16
        add     ecx,16
        dec     edx
        jnz     _SCR0
        ret
EndP SCRebuild

; *****************************************************************************
;
;       Display file execute. PC is now in the range 8000-FFFF.
;
;       Instructions should be executed if they are in the 6th bit
;       of their Opcode. If as a result of this execution PC
;       drops back below 8000, return from caller.
;
;       Otherwise instructions are 'displayed'. (e.g. it is char data)
;
;       If bit 6 of R goes low, then an interrupt (CheckInterrupt)
;       will take place. 
;
; *****************************************************************************

SCExecDisplayFile Proc

        mov     eax,[VideoAddress]      ; set the rendering pointer
        mov     [RenderPointer],eax     ; for this line

        cmp     [Redraw],0              ; skip if NO redraw
        jnz     _EDF0                   ; this part,which sets up & clears
        cmp     [VideoLine],193         ; also, if past the vertical end
        ja      _EDF0                   ; of the screen
      
        mov     ebx,[Background]        ; clear the background
        mov     ecx,64                  ; 256 bytes,32 chars
_EDFCl: mov     [eax],ebx
        add     eax,4
        dec     ecx
        jnz     _EDFCl

;
;       MAIN EXECUTE DISPLAY LOOP
;
_EDF0:  test    [byte ptr _PC+1],080h   ; still in the range 8000-8FFF
        jz      _EDFX                   ; if not, we can Exit !!!!

        mov     ebx,[dword ptr _PC]     ; address to read
        and     ebx,07FFFh              ; drop bit 15 because it's dfile.
        mov     al,[byte ptr Memory+ebx]; read the opcode there
        test    al,64                   ; is it a displayable character ?
        jz      _EDF1                   ; yes, then display it

        call    XEcuteAlways            ; Execute the command.

_EDF2:                                  ; return to this point
        test    [_R],64                 ; is bit 6 or R zero ?
        jnz     _EDF0                   ; if not, loop back

        call    CheckInterrupt          ; if it is, then look at IRQ ?
        jmp     _EDF0                   ; and loop back

_EDFX:  cmp     [ULADelay],0            ; Delay required for 02B5 call
        jnz     _EDFX1                  ; does a short line....

        inc     [ULALine]               ; bump the ULA Line counter
        and     [ULALine],7             ; it is a 3 bit counter
_EDFX1: mov     [ULADelay],0            ; don't need the delay any more
        cmp     [NMIVSCounter],VBlank   ; are we in top vblank
        jb      _EDFX2                  ; if so don't move the pointer
        inc     [VideoLine]             ; increment the video line counter
        add     [VideoAddress],320      ; next video line
_EDFX2:
        ret
;
;       Displayable character in DFILE (bit 6 is zero)
;
_EDF1:  xor     ebx,ebx                 ; get for HiRes test
        xor     ecx,ecx                 ; CL = 0 and EBX = Address
        mov     bh,[byte ptr _I]        ; build address out of I and R
        mov     bl,[byte ptr _R]
        and     bl,07Fh
        and     [byte ptr _RBit7],080h  ; R is a 7 bit counter with a
        or      bl,[byte ptr _RBit7]    ; single data bit.

        inc     [byte ptr _R]           ; increment the refresh register
        and     [byte ptr _R],127       ; mask 7 bits
        inc     [word ptr _PC]          ; bump the program counter

        cmp     [Redraw],0              ; redrawing ?
        jnz     _EDF2
        cmp     [VideoLine],193         ; off screen ?
        ja      _EDF2

        cmp     bx,02000h               ; is the I register pointing to RAM?
        jae     _EDFHiRes               ; if so, goto HiRes code

        mov     cl,al                   ; save character for inverse test
        and     al,63                   ; drop bits 6 and 7

        xor     ebx,ebx                 ; make an offset into the charset
        mov     bl,al                   ; bits 3..8 from the character
        shl     ebx,3                   ; in AL, put it in the right place.

        mov     dl,[_I]                 ; use Bits 7-1 of I only
        and     dl,0FEh                 ; why Zombies doesn't work on xtender
        or      bh,dl                   ; bits 8..15 are in 'I'

        or      bl,[byte ptr ULALine]   ; other bits from ULA Line counter

        mov     al,[byte ptr Memory+ebx]; read character ROM.
        test    cl,128                  ; inverse ?
        jz      _EDF10
        xor     al,255                  ; invert all the bits
        jmp     _EDF10

_EDFHiRes:                              ; Hi res comes here, we just read
        mov     al,[byte ptr Memory+ebx]; memory !
_EDF10:
        mov     ebx,[RenderPointer]     ; it goes at ebx
        mov     dh,[byte ptr Foreground]; in this colour

_EDF11: test    al,128                  ; render AL to EBX. Can be optimised!
        jz      _EDF12
        mov     [ebx],dh
_EDF12: inc     ebx
        add     al,al
        jnz     _EDF11
        add     [RenderPointer],8

        jmp     _EDF2                   ; go back, finish and do next char

EndP SCExecDisplayFile

; *****************************************************************************
;
;                                 Fast Paint Hack
;
; *****************************************************************************

SCFastPaint Proc
;       jmp     DefaultCode             ; take this out and paint via DFILE
                                        ; using the Real Int Code. (slower!)

        cmp     [byte ptr _C],8         ; C must be 8 - indicates whole char
        jne     DefaultCode
        cmp     [VideoLine],186         ; must still be on the screen
        jae     DefaultCode
        cmp     [word ptr _IX],02000h   ; IX must point to ROM
        jae     DefaultCode
        mov     ebx,[dword ptr _HL]     ; check first byte not code
        and     ebx,07FFFh              ; for Erno Borgesteede's DFILE patch
        mov     al,[byte ptr Memory+ebx]
        test    al,64
        jnz     DefaultCode
        cmp     [IsNoAccelerate],0      ; no accleration ?
        jnz     DefaultCode

        mov     [byte ptr _C],1         ; set C to 1 for a quick return.
                                        ; we do 8 rows at a time so RST38 is
                                        ; only called about 25 times/frame.

        call    _OG0_FB                 ; enable the interrupts

        cmp     [Redraw],0              ; don't redraw ?
        jnz     _NoRedraw

        mov     ebx,[VideoAddress]      ; clear the background
        mov     eax,[Background]
        mov     ecx,256/4
_FP2:   mov     [ebx],eax
        mov     [ebx+320],eax
        mov     [ebx+640],eax
        mov     [ebx+960],eax
        mov     [ebx+1280],eax
        mov     [ebx+1600],eax
        mov     [ebx+1920],eax
        mov     [ebx+2240],eax
        add     ebx,4
        dec     ecx
        jnz     _FP2

        mov     ebx,[VideoAddress]      ; now render the characters
        mov     ebp,[dword ptr _HL]     ; address of the characters
        and     ebp,07FFFh              ; drop bit 15
        add     ebp,offset Memory       ; in real memory now.
_FP3:   mov     al,[ebp]                ; look at the character
        cmp     al,118
        je      _FP4                    ; exit if halt found
        call    _FPRender
        add     ebx,8
        inc     ebp                     ; do the next character
        jmp     _FP3
_FP4:

_NoRedraw:                              ; finish up.....
        mov     ebx,[dword ptr _HL]     ; skip forward to the halt
        and     ebx,07FFFh              ; we're in HIRAM here
_FP0:   cmp     [byte ptr Memory+ebx],118
        je      _FP1
        inc     ebx
        jmp     _FP0

_FP1:   or      ebx,08000h              ; the address of the HALT 'virtual'
        mov     [word ptr _PC],bx       ; update the address
        cmp     [NMIVSCounter],VBlank   ; update pointers only if not
        jb      _FP5                    ; in VBLANK
        add     [VideoLine],8           ; doing lines in groups of 8.
        add     [VideoAddress],320*8
_FP5:   call    CheckInterrupt          ; lets have an interrupt....
        ret                             ; exit

_FPRender:
        push    ebx

        mov     cl,al                   ; save character in CL
        and     cl,128                  ; if & $80 non zero make it 255
        jz      _FPR0a                  ; this is to be xored later on
        mov     cl,255                  ; to do the inverting
_FPR0a:   

        and     eax,63                  ; mask off bits 6 and 7
        shl     eax,3                   ; al = al * 8
        mov     dl,[_I]                 ; use 7 bits of I only
        and     dl,0FEh
        or      ah,dl                   ; eax now is virtual address of table
        add     eax,offset Memory       ; eax points to real memory data
        mov     esi,eax                 ; put character data in ESI

        mov     ch,8                    ; ch = characters to write
        mov     dh,[byte ptr Foreground]; dl = foreground colour

_FPR0:  push    ebx                     ; save start of line
        mov     al,[esi]                ; get character from CGEN space
        xor     al,cl
        inc     esi                     ; point to next character

_FPR1:  test    al,128                  ; render the pattern to video ram
        jz      _FPR2
        mov     [ebx],dh
_FPR2:  inc     ebx
        add     al,al
        jnz     _FPR1
        pop     ebx                     ; restore start of line

        add     ebx,320                 ; next vertical line
        dec     ch
        jnz     _FPR0                   ; do for all 8 chars
                           
        pop     ebx
        ret
;
;       Coming here executes the instructions we patched out.
;       by overwriting 43,44 they are EI JP(HL)
;
DefaultCode:
        call    _OG0_FB                 ; enable interrupts
        call    _OG0_E9                 ; jp (HL)
        ret
EndP SCFastPaint

.data?                                  ; buffer for Video RAM
VBuffer db 65530 dup (?)
VBuffEnd db 20480 dup (?)

.code

; *****************************************************************************
;
;                           Update Colour Scheme
;
; *****************************************************************************

SCResetBackground Proc
        mov     eax,[Scheme]
        mov     ebx,[SchTable+eax*8]
        mov     [Foreground],ebx
        mov     ebx,[SchTable+eax*8+4]
        mov     [Background],ebx
        call    SCFrame
        ret
.data
SchTable dd     00000000h,0f0f0f0fh     ; Colour Schemes
        dd      0f0f0f0fh,00000000h
        dd      0A0A0A0Ah,00000000h
        dd      00000000h,0b0b0b0bh
.code
EndP SCResetBackground

; *****************************************************************************
;
;                       Used to emulate 'Fast' Mode
;
; *****************************************************************************

SCFastMode Proc
        cmp     [InFastMode],0          ; if going into fast mode
        jnz     _SCFM2
        mov     [InFastMode],1
        mov     ebx,[VRAMPtr]           ; completely switch the screen off
        mov     ecx,320*200/4
        xor     eax,eax
_SCFM0: mov     [ebx],eax
        add     ebx,4
        dec     ecx
        jnz     _SCFM0
_SCFM2: ret
EndP SCFastMode
end
